from batou.update import update_bootstrap
from batou.utils import cmd
from batou import output
import argparse
import os.path
import pkg_resources
import shutil


def main(destination):
    develop = os.environ['BATOU_DEVELOP']
    if develop:
        output.warn(
            'Initializing with a development copy of batou will cause your '
            'project to have a reference outside its repository. '
            'Use at your own risk. ')
        develop = os.path.abspath(develop)
    print('Bootstrapping new batou project in {}. This can take a while.'
          .format(os.path.abspath(destination)))
    if os.path.exists(destination):
        print('{} exists already. Not copying template structure.'.format(
              destination))
        os.chdir(destination)
    else:
        source = os.path.dirname(__file__) + '/init-template'
        shutil.copytree(source, destination)
        os.chdir(destination)
        cmd('hg -y init .')
    update_bootstrap(os.environ['BATOU_VERSION'], develop)
    # Need to clean up to avoid inheriting info that we're bootstrapped
    # already.
    for key in list(os.environ):
        if key.startswith('BATOU_'):
            del os.environ[key]
    cmd('./batou --help')


def console_main():
    parser = argparse.ArgumentParser(
        description="""\
Initialize batou project in the given directory. If the given directory does
not exist, it will be created.

If no directory is given, the current directory is used.
""")
    parser.add_argument('destination')

    os.environ['BATOU_VERSION'] = pkg_resources.require('batou')[0].version
    os.environ['BATOU_DEVELOP'] = ''
    args = parser.parse_args()
    main(args.destination)



from .environment import Environment, MissingEnvironment
from .utils import locked, self_id
from .utils import notify
from batou import DeploymentError, ConfigurationError
from batou._output import output, TerminalBackend
import sys
import threading


class Connector(threading.Thread):

    def __init__(self, host):
        self.host = host
        self.exc_info = None
        super(Connector, self).__init__(name=host.name)

    def run(self):
        try:
            self.host.connect()
            self.host.start()
        except Exception:
            self.exc_info = sys.exc_info()

    def join(self):
        super(Connector, self).join()
        if self.exc_info:
            raise self.exc_info[0], self.exc_info[1], self.exc_info[2]


class Deployment(object):

    _upstream = None

    def __init__(self, environment, platform, timeout, dirty, fast,
                 reset=False):
        self.environment = environment
        self.platform = platform
        self.timeout = timeout
        self.dirty = dirty
        self.fast = fast
        self.reset = reset

    def load(self):
        output.section("Preparing")

        output.step("main",
                    "Loading environment `{}`...".format(self.environment))

        self.environment = Environment(
            self.environment, self.timeout, self.platform)
        self.environment.deployment = self
        self.environment.load()

        # This is located here to avoid duplicating the verification check
        # when loading the repository on the remote environment object.
        output.step("main", "Verifying repository ...")
        self.environment.repository.verify()

        output.step("main", "Loading secrets ...")
        self.environment.load_secrets()

    def configure(self):
        output.section("Configuring first host")
        self.connections = iter(self._connections())
        self.connections.next().join()

    def _connections(self):
        self.environment.prepare_connect()
        for i, host in enumerate(self.environment.hosts.values(), 1):
            if host.ignore:
                output.step(host.name, "Connection ignored ({}/{})".format(
                    i, len(self.environment.hosts)),
                    bold=False, red=True)
                continue
            output.step(host.name, "Connecting via {} ({}/{})".format(
                        self.environment.connect_method, i,
                        len(self.environment.hosts)))
            c = Connector(host)
            c.start()
            yield c

    def connect(self):
        output.section("Connecting remaining hosts")
        # Consume the connection iterator to establish remaining connections.
        [t.join() for t in list(self.connections)]

    def deploy(self, predict_only=False):
        if predict_only:
            output.section("Predicting deployment actions")
        else:
            output.section("Deploying")

        # Pick a reference remote (the last we initialised) that will pass us
        # the order we should be deploying components in.
        reference_node = [h for h in self.environment.hosts.values()
                          if not h.ignore][0]

        for root in reference_node.roots_in_order():
            hostname, component, ignore_component = root
            host = self.environment.hosts[hostname]
            if host.ignore:
                output.step(
                    hostname,
                    "Skipping component {} ... (Host ignored)".format(
                        component), red=True)
                continue
            if ignore_component:
                output.step(
                    hostname, "Skipping component {} ... (Component ignored)".
                    format(component), red=True)
                continue

            output.step(
                hostname, "Deploying component {} ...".format(component))
            host.deploy_component(component, predict_only)

    def disconnect(self):
        output.step("main", "Disconnecting from nodes ...", debug=True)
        for node in self.environment.hosts.values():
            node.disconnect()


def main(environment, platform, timeout, dirty, fast, consistency_only,
         predict_only, reset):
    output.backend = TerminalBackend()
    output.line(self_id())
    if consistency_only:
        ACTION = 'CONSISTENCY CHECK'
    elif predict_only:
        ACTION = 'DEPLOYMENT PREDICTION'
    else:
        ACTION = 'DEPLOYMENT'
    with locked('.batou-lock'):
        try:
            deployment = Deployment(
                environment, platform, timeout, dirty, fast,
                reset=reset)
            deployment.load()
            deployment.configure()
            if not consistency_only:
                deployment.connect()
                deployment.deploy(predict_only)
            deployment.disconnect()
        except MissingEnvironment as e:
            e.report()
            output.section("{} FAILED".format(ACTION), red=True)
            notify('{} FAILED'.format(ACTION),
                   'Configuration for {} encountered an error.'.format(
                       environment))
            sys.exit(1)
        except ConfigurationError as e:
            output.section("{} FAILED".format(ACTION), red=True)
            notify('{} FAILED'.format(ACTION),
                   'Configuration for {} encountered an error.'.format(
                       environment))
            sys.exit(1)
        except DeploymentError as e:
            e.report()
            output.section("{} FAILED".format(ACTION), red=True)
            notify('{} FAILED'.format(ACTION),
                   '{} encountered an error.'.format(environment))
            sys.exit(1)
        except Exception:
            # An unexpected exception happened. Bad.
            output.error("Unexpected exception", exc_info=sys.exc_info())
            output.section("{} FAILED".format(ACTION), red=True)
            notify('{} FAILED'.format(ACTION),
                   'Encountered an unexpected exception.')
            sys.exit(1)
        else:
            output.section('{} FINISHED'.format(ACTION), green=True)
            notify('{} SUCCEEDED'.format(ACTION), environment)



from batou.environment import Environment
from batou.deploy import Deployment
from batou.host import RemoteHost
from batou.utils import cmd
import pytest
import mock


@pytest.mark.slow
def test_remote_deployment_initializable(sample_service):
    cmd('hg init')
    with open('.hg/hgrc', 'w') as f:
        f.write('[paths]\ndefault=https://example.com')
    env = Environment('test-with-env-config')
    env.load()
    env.configure()
    Deployment(env, platform='', timeout=30, fast=False, dirty=False)


def test_remote_bundle_breaks_on_missing_head(sample_service):
    cmd('hg init')
    env = mock.Mock()
    env.base_dir = sample_service
    h = RemoteHost('asdf', env)
    from batou.repository import MercurialBundleRepository
    repository = MercurialBundleRepository(env)
    h.rpc = mock.Mock()
    h.rpc.hg_current_heads.return_value = []
    with pytest.raises(ValueError) as e:
        repository.update(h)
    assert e.value.message == (
        'Remote repository did not find any heads. '
        'Can not continue creating a bundle.')



import argparse
import batou
import batou.deploy
import batou.init
import batou.secrets.edit
import batou.secrets.manage
import batou.update
import os
import os.path
import pkg_resources
import sys


def main():
    os.chdir(os.path.dirname(sys.argv[0]))
    parser = argparse.ArgumentParser(
        description='batou v%s '
        'multi-(host|component|environment|version|platform) deployment'
        % pkg_resources.resource_string(__name__, 'version.txt'),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-d', '--debug', action='store_true',
        help='Enable debug mode.')
    parser.add_argument(
        '-F', '--fast', action='store_true',
        help='Enable fast mode. Do not perform bootstrapping.')
    parser.add_argument(
        '--reset', action='store_true',
        help='Reset batou environment.')

    subparsers = parser.add_subparsers()

    # Deploy
    p = subparsers.add_parser(
        'deploy', help=u'Deploy an environment.')
    p.add_argument(
        '-p', '--platform', default=None,
        help='Alternative platform to choose. Empty for no platform.')
    p.add_argument(
        '-t', '--timeout', default=None,
        help='Override the environment\'s timeout setting')
    p.add_argument(
        '-D', '--dirty', action='store_true',
        help='Allow deploying with dirty working copy or outgoing changes.')
    p.add_argument(
        '-c', '--consistency-only', action='store_true',
        help='Only perform a deployment model and environment '
             'consistency check. Only connects to a single host. '
             'Does not touch anything.')
    p.add_argument(
        '-P', '--predict-only', action='store_true',
        help='Only predict what updates would happen. '
             'Do not change anything.')
    p.add_argument(
        'environment', help='Environment to deploy.',
        type=lambda x: x.replace('.cfg', ''))
    p.set_defaults(func=batou.deploy.main)

    # SECRETS
    secrets = subparsers.add_parser(
        'secrets', help="""\
Manage encrypted secret files.

        Relies on gpg being installed and configured correctly.
""")

    sp = secrets.add_subparsers()

    p = sp.add_parser('edit',
                      help="""\
Encrypted secrets file editor utility. Decrypts file,
invokes the editor, and encrypts the file again. If called with a
non-existent file name, a new encrypted file is created.
""")
    p.add_argument('--editor', '-e', metavar='EDITOR',
                   default=os.environ.get('EDITOR', 'vi'),
                   help='Invoke EDITOR to edit (default: $EDITOR or vi)')
    p.add_argument(
        'environment', help='Environment to edit secrets for.',
        type=lambda x: x.replace('.cfg', ''))
    p.set_defaults(func=batou.secrets.edit.main)

    p = sp.add_parser('summary',
                      help="""\
Give a summary of secret files and who has access.
""")
    p.set_defaults(func=batou.secrets.manage.summary)

    p = sp.add_parser('add',
                      help="""\
Add a user's key to one or more secret files.
""")
    p.add_argument(
        'keyid', help='The user\'s key ID or email address')
    p.add_argument(
        '--environments',
        default='',
        help='The environments to update. Update all if not specified.')
    p.set_defaults(func=batou.secrets.manage.add_user)

    p = sp.add_parser('remove',
                      help="""\
Remove a user's key from one or more secret files.
""")
    p.add_argument(
        'keyid', help='The user\'s key ID or email address')
    p.add_argument(
        '--environments',
        default='',
        help='The environments to update. Update all if not specified.')
    p.set_defaults(func=batou.secrets.manage.remove_user)

    # INIT
    p = subparsers.add_parser(
        'init',
        help="""\
Initialize batou project in the given directory. If the given directory does
not exist, it will be created.

If no directory is given, the current directory is used.
""")
    p.add_argument('destination')
    p.set_defaults(func=batou.init.main)

    # UPDATE
    p = subparsers.add_parser(
        'update', help=u'Update the batou version.')
    group = p.add_mutually_exclusive_group(required=True)
    group.add_argument(
        '--version', help='Exact version to install.',
        default='')
    group.add_argument(
        '--develop', help='Path to checkout of batou to install in edit mode.',
        default='')
    p.add_argument(
        '--finish', help='(internal)', action='store_true')
    p.set_defaults(func=batou.update.main)

    args = parser.parse_args()

    # Consume global arguments
    batou.output.enable_debug = args.debug

    # Pass over to function
    func_args = dict(args._get_kwargs())
    del func_args['func']
    del func_args['debug']
    args.func(**func_args)



import batou.component
import batou.lib.service
import pytest


@pytest.fixture
def deploy_platform():
    return 'test'


@pytest.yield_fixture
def service(root, tmpdir, request, deploy_platform):
    assert not hasattr(batou.lib.service.Service, '_platforms')
    root.host.platform = deploy_platform

    @batou.component.platform('test', batou.lib.service.Service)
    class TestPlatformService(batou.component.Component):

        def start(self):
            service._platform_started = True

    @batou.component.platform('nostart', batou.lib.service.Service)
    class NoStartPlatformService(batou.component.Component):
        pass

    batou.lib.service.Service._add_platform('test', TestPlatformService)

    service = batou.lib.service.Service(
        'touch {}/called'.format(tmpdir))
    service._platform_started = False
    root.component += service
    root.component.deploy()
    yield service
    del batou.lib.service.Service._platforms


def test_start_delegates_to_platform_component(service, tmpdir):
    service.start()
    # The platform specific start() was called
    assert service._platform_started
    # The default implementation was not called
    assert not tmpdir.join('called').check()


@pytest.mark.parametrize('deploy_platform', ['another'])
def test_start_runs_executable_if_no_platform_component(service, tmpdir):
    service.start()
    assert tmpdir.join('called').check()


@pytest.mark.parametrize('deploy_platform', ['nostart'])
def test_start_runs_executable_if_no_platform_specific_start(service, tmpdir):
    service.start()
    assert tmpdir.join('called').check()



from batou.utils import cmd
import batou.template
import os.path


bootstrap_template = os.path.dirname(__file__) + '/bootstrap-template'


def generate_bootstrap(version, develop):
    engine = batou.template.Jinja2Engine()
    template = open(bootstrap_template, 'r').read()
    return engine.expand(template, dict(version=version, develop=develop))


def update_bootstrap(version, develop):
    bootstrap = generate_bootstrap(version, develop)
    with open('batou', 'w') as b:
        b.write(bootstrap)
    os.chmod('batou', 0o755)


def main(version, develop, finish, fast=False):
    # Record that we're updating. This will be called twice: first when the
    # user asks for updating, which ensures that any subsequent calls won't
    # downgrade in the first bootstrap phase. Second, we call the newly
    # installed batou with '--finish' to update the early bootstrapping file
    # itself.
    update_bootstrap(version, develop)
    if finish:
        return
    if version:
        print "Updating batou to {}".format(version)
        print (
            "See the changelog at "
            "https://batou.readthedocs.org/en/latest/changes.html")
        cmd('.batou/bin/pip install --egg batou=={}'.format(version))
    else:
        cmd('.batou/bin/pip install --egg -e {}'.format(develop))
    cmd('./batou update --version=\'{}\' '
        '\'--develop={}\' --finish'.format(version, develop))



"""Securely edit encrypted secret files."""

from .encryption import EncryptedConfigFile
from batou.lib.file import ensure_path_nonexistent
import os
import subprocess
import sys
import tempfile


class Editor(object):

    def __init__(self, editor, encrypted_file):
        self.editor = editor
        self.encrypted_file = encrypted_file
        self.original_cleartext = encrypted_file.read()
        self.cleartext = self.original_cleartext

    def main(self):
        cmd = 'edit'
        while cmd != 'quit':
            try:
                self.process_cmd(cmd)
            except Exception as e:
                print
                print "Could not update due to error: {}".format(e)
                print "Your changes are still available. You can try:"
                print "\tedit (opens editor with current data again)"
                print "\tencrypt (tries to encrypt current data again)"
                print "\tquit (quits and loses your changes)"
                cmd = raw_input("> ").strip()
            else:
                break

    def process_cmd(self, cmd):
            if cmd == 'edit':
                self.edit()
                self.encrypt()
            elif cmd == 'encrypt':
                self.encrypt()
            else:
                print "Did not understand command '{}'".format(cmd)

    def encrypt(self):
        if self.cleartext == self.original_cleartext:
            print "No changes from original cleartext. Not updating."
            return
        self.encrypted_file.write(self.cleartext)

    def edit(self):
        with tempfile.NamedTemporaryFile(
                prefix='edit', suffix='.cfg') as clearfile:
            clearfile.write(self.cleartext)
            clearfile.flush()

            subprocess.check_call(
                [self.editor + ' ' + clearfile.name],
                shell=True)

            with open(clearfile.name, 'r') as new_clearfile:
                self.cleartext = new_clearfile.read()


def main(editor, environment, fast):
    """Secrets editor console script.

    The main focus here is to avoid having unencrypted files accidentally
    ending up in the deployment repository.

    """
    encrypted = 'secrets/{}.cfg'.format(environment)

    if not os.path.exists('environments/{}.cfg'.format(environment)):
        print "Environment '{}' does not exist. Typo?".format(environment)
        print "Existing environments:"
        print "\n".join(os.listdir('environments')).replace('.cfg', '')
        sys.exit(1)

    if not os.path.isdir('secrets'):
        ensure_path_nonexistent('secrets')
        os.mkdir('secrets')

    with EncryptedConfigFile(encrypted, write_lock=True) as sf:
        editor = Editor(editor, sf)
        editor.main()



from batou.component import Component
from batou import UpdateNeeded


class DPKG(Component):
    """Install a dpkg system package.

    This is experimental to get automation started, apt only for now.

    """

    namevar = 'package'

    def verify(self):
        stdout, stderr = self.cmd('LANG=C dpkg --get-selections')
        for line in stdout.splitlines():
            candidate, status = line.split()
            if candidate == self.package and status == 'install':
                break
        else:
            raise UpdateNeeded()

    def update(self):
        # /dev/null: a talky apt-get get blocks remote execution
        self.cmd('LANG=C apt-get -qy install {0}'.format(self.package))



from batou.environment import Environment
from batou.utils import cmd
import os
import os.path
from batou.tests.ellipsis import Ellipsis


def test_service_early_resource():
    env = Environment(
        'dev',
        basedir=os.path.dirname(__file__) + '/fixture/service_early_resource')
    env.load()
    env.configure()
    assert env.resources.get('zeo') == ['127.0.0.1:9000']


def test_example_errors():
    os.chdir('examples/errors')
    out, _ = cmd('./batou deploy errors', acceptable_returncodes=[1])
    assert out == Ellipsis("""\
batou/... (CPython 2.7..., ...)
================================== Preparing =================================\
==
main: Loading environment `errors`...
main: Verifying repository ...
main: Loading secrets ...
============================ Configuring first host ==========================\
==
localhost: Connecting via local (1/1)
ERROR: Failed loading component file
           File: .../component5/component.py
      Exception: invalid syntax (component.py, line 1)
ERROR: Failed loading component file
           File: .../component6/component.py
      Exception: No module named asdf
ERROR: Missing component
      Component: missingcomponent
           Host: localhost
ERROR: Superfluous section in environment configuration
        Section: superfluoussection
ERROR: Override section for unknown component found
      Component: nonexisting-component-section
ERROR: crontab@localhost: No cron jobs found.
ERROR: Failed override attribute conversion
           Host: localhost
      Attribute: Component1.do_what_is_needed
     Conversion: convert_literal(u'false')
          Error: malformed string
ERROR: Overrides for undefined attributes
           Host: localhost
      Component: Component2
     Attributes: this_does_not_exist
ERROR: Failed override attribute conversion
           Host: localhost
      Attribute: DNSProblem.attribute_with_problem
     Conversion: Address(u'localhost')
          Error: Need port for service address.
ERROR: Unused provided resources
        backend: ['192.168.0.1']
       frontend: ['test00.gocept.net']
            sub: [<SubComponent (localhost) "Component3 > SubComponent(sub sub)">]
ERROR: Unsatisfied resource requirements
     application: <undefined>
ERROR: Found dependency cycle
     cycle1 depends on
             cycle2
     cycle2 depends on
             cycle1
ERROR: 6 remaining unconfigured component(s)
======================= 13 ERRORS - CONFIGURATION FAILED =====================\
==
============================== DEPLOYMENT FAILED =============================\
==
""")  # NOQA


def test_example_errors_missing_environment():
    os.chdir('examples/errors')
    out, _ = cmd('./batou deploy production', acceptable_returncodes=[1])
    assert out == Ellipsis("""\
batou/... (CPython 2.7..., ...)
================================== Preparing =================================\
==
main: Loading environment `production`...
ERROR: Missing environment
     Environment: production
============================== DEPLOYMENT FAILED =============================\
==
""")  # NOQA


def test_example_ignores():
    os.chdir('examples/ignores')
    out, _ = cmd('./batou deploy ignores')
    assert out == Ellipsis("""\
batou/... (CPython 2.7..., ...)
================================== Preparing =================================\
==
main: Loading environment `ignores`...
main: Verifying repository ...
main: Loading secrets ...
============================ Configuring first host ==========================\
==
otherhost: Connection ignored (1/2)
localhost: Connecting via local (2/2)
========================== Connecting remaining hosts ========================\
==
================================== Deploying =================================\
==
localhost: Deploying component component1 ...
localhost: Skipping component fail ... (Component ignored)
otherhost: Skipping component fail2 ... (Host ignored)
============================= DEPLOYMENT FINISHED ============================\
==
""")



from batou.component import Component
from batou.lib.archive import Extract
from batou.lib.download import Download
import os.path


class Configure(Component):
    # XXX not convergent when changing args

    namevar = 'path'
    args = ''
    prefix = None
    build_environment = None

    def configure(self):
        if self.prefix is None:
            self.prefix = self.workdir

    def verify(self):
        # This is guesswork. Unfortunately CMMI doesn't work any better.
        self.assert_file_is_current(self.path + '/config.status')
        self.assert_file_is_current(
            self.path + '/.batou.config.success',
            [self.path + '/configure'])

    def update(self):
        with self.chdir(self.path):
            self.cmd(self.expand(
                './configure --prefix={{component.prefix}} '
                '{{component.args}}'), env=self.build_environment)
            self.touch('.batou.config.success')


class Make(Component):

    namevar = 'path'
    build_environment = None

    def verify(self):
        self.assert_file_is_current(
            self.path + '/.batou.make.success', [self.path + '/Makefile'])

    def update(self):
        with self.chdir(self.path):
            self.cmd('make install', env=self.build_environment)
            self.touch('.batou.make.success')


class Build(Component):
    """Complex build definition for

    - downloading
    - extracting
    - configure
    - make install

    """

    namevar = 'uri'
    checksum = None
    configure_args = ''
    prefix = None
    build_environment = None

    def configure(self):
        download = Download(
            self.uri, checksum=self.checksum)
        self += download

        extract = Extract(download.target, strip=1)
        self += extract

        self += Configure(extract.target,
                          args=self.configure_args, prefix=self.prefix,
                          build_environment=self.build_environment)
        self += Make(extract.target, build_environment=self.build_environment)

    @property
    def namevar_for_breadcrumb(self):
        return os.path.basename(self.uri)



from batou import DeploymentError, output, RepositoryDifferentError
from batou.utils import cmd as cmd_, CmdExecutionError
import execnet
import os
import subprocess
import tempfile


def cmd(c, *args, **kw):
    return cmd_('LANG=C LC_ALL=C LANGUAGE=C {}'.format(c), *args, **kw)


def find_line_with(prefix, output):
    for line in output.splitlines():
        line = line.strip()
        if line.startswith(prefix):
            return line.replace(prefix, '', 1).strip()


class Repository(object):
    """A repository containing the batou deployment.

    The actual deployment may be located within a prefix
    of this repository. Where the repository starts can be
    determined by the specific repository implementation.

    """

    root = '.'

    def __init__(self, environment):
        self.environment = environment

    @classmethod
    def from_environment(cls, environment):
        if environment.connect_method == 'local':
            return NullRepository(environment)
        elif environment.update_method == 'rsync':
            return RSyncRepository(environment)
        elif environment.update_method == 'hg-bundle':
            return MercurialBundleRepository(environment)
        elif environment.update_method == 'hg-pull':
            return MercurialPullRepository(environment)
        elif environment.update_method == 'git-bundle':
            return GitBundleRepository(environment)
        elif environment.update_method == 'git-pull':
            return GitPullRepository(environment)
        raise ValueError('Could not find method to transfer the repository.')

    def verify(self):
        pass

    def update(self):
        pass


class NullRepository(Repository):
    """A repository that does nothing to verify or update."""


class RSyncRepository(Repository):

    root = '.'

    def verify(self):
        output.annotate(
            "You are using rsync. This is a non-verifying repository "
            "-- continuing on your own risk!", red=True)

    def update(self, host):
        env = self.environment
        blacklist = ['.batou', 'work', '.git', '.hg', '.vagrant',
                     '.batou-lock']
        for candidate in os.listdir(env.base_dir):
            if candidate in blacklist:
                continue

            source = os.path.join(env.base_dir, candidate)
            target = os.path.join(host.remote_base, candidate)
            output.annotate("rsync: {} -> {}".format(source, target),
                            debug=True)
            rsync = execnet.RSync(source, verbose=False)
            rsync.add_target(host.gateway, target)
            rsync.send()


class MercurialRepository(Repository):

    root = None
    _upstream = None

    def __init__(self, environment):
        super(MercurialRepository, self).__init__(environment)
        self.root = subprocess.check_output(['hg', 'root']).strip()
        self.branch = environment.branch or 'default'
        self.subdir = os.path.relpath(
            self.environment.base_dir, self.root)

    @property
    def upstream(self):
        if self.environment.repository_url is not None:
            self._upstream = self.environment.repository_url
        elif self._upstream is None:
            self._upstream = cmd('hg showconfig paths')[0]
            self._upstream = self._upstream.split('\n')[0].strip()
            assert self._upstream.startswith('paths.default')
            self._upstream = self.upstream.split('=')[1]
        return self._upstream

    def update(self, host):
        self._ship(host)
        remote_id = host.rpc.hg_update_working_copy(self.branch)
        local_id, _ = cmd('hg id -i')
        if self.environment.deployment.dirty:
            local_id = local_id.replace('+', '')
        local_id = local_id.strip()
        if remote_id != local_id:
            raise RepositoryDifferentError(local_id, remote_id)

    def verify(self):
        # Safety belt that we're acting on a clean repository.
        if self.environment.deployment.dirty:
            output.annotate(
                "You are running a dirty deployment. This can cause "
                "inconsistencies -- continuing on your own risk!", red=True)
            return

        try:
            status, _ = cmd('hg -q stat')
        except CmdExecutionError:
            output.error('Unable to check repository status. '
                         'Is there an HG repository here?')
            raise
        else:
            status = status.strip()
            if status.strip():
                output.error("Your repository has uncommitted changes.")
                output.annotate("""\
I am refusing to deploy in this situation as the results will be unpredictable.
Please commit and push first.
""", red=True)
                output.annotate(status, red=True)
                raise DeploymentError()
        try:
            cmd('hg -q outgoing -l 1', acceptable_returncodes=[1])
        except CmdExecutionError:
            output.error("""\
Your repository has outgoing changes.

I am refusing to deploy in this situation as the results will be unpredictable.
Please push first.
""")
            raise DeploymentError()


class MercurialPullRepository(MercurialRepository):

    def _ship(self, host):
        host.rpc.hg_pull_code(upstream=self.upstream)


class MercurialBundleRepository(MercurialRepository):

    def _ship(self, host):
        heads = host.rpc.hg_current_heads()
        if not heads:
            raise ValueError("Remote repository did not find any heads. "
                             "Can not continue creating a bundle.")
        fd, bundle_file = tempfile.mkstemp()
        os.close(fd)
        bases = ' '.join('--base {}'.format(x) for x in heads)
        cmd('hg -qy bundle {} {}'.format(bases, bundle_file),
            acceptable_returncodes=[0, 1])
        change_size = os.stat(bundle_file).st_size
        if not change_size:
            return
        output.annotate(
            'Sending {} bytes of changes'.format(change_size), debug=True)
        rsync = execnet.RSync(bundle_file, verbose=False)
        rsync.add_target(host.gateway,
                         host.remote_repository + '/batou-bundle.hg')
        rsync.send()
        os.unlink(bundle_file)
        output.annotate(
            'Unbundling changes', debug=True)
        host.rpc.hg_unbundle_code()


class GitRepository(Repository):

    root = None
    _upstream = None
    remote = 'origin'

    def __init__(self, environment):
        super(GitRepository, self).__init__(environment)
        self.branch = environment.branch or 'master'
        self.root = subprocess.check_output(
            ['git', 'rev-parse', '--show-toplevel']).strip()
        self.subdir = os.path.relpath(
            self.environment.base_dir, self.root)

    @property
    def upstream(self):
        if self.environment.repository_url is not None:
            self._upstream = self.environment.repository_url
        elif self._upstream is None:
            result = cmd('git remote show -n {}'.format(self.remote))[0]
            self._upstream = find_line_with('Fetch URL:', result)
        return self._upstream

    def update(self, host):
        self._ship(host)
        remote_id = host.rpc.git_update_working_copy(self.branch)
        # This can theoretically fail if we have a fresh repository, but
        # that doesn't make sense at this point anyway.
        local_id, _ = cmd('git rev-parse HEAD')
        local_id = local_id.strip()
        if remote_id != local_id:
            raise RepositoryDifferentError(local_id, remote_id)

    def verify(self):
        # Safety belt that we're acting on a clean repository.
        if self.environment.deployment.dirty:
            output.annotate(
                "You are running a dirty deployment. This can cause "
                "inconsistencies -- continuing on your own risk!", red=True)
            return

        try:
            status, _ = cmd('git status --porcelain')
        except CmdExecutionError:
            output.error('Unable to check repository status. '
                         'Is there a Git repository here?')
            raise
        else:
            status = status.strip()
            if status.strip():
                output.error("Your repository has uncommitted changes.")
                output.annotate("""\
I am refusing to deploy in this situation as the results will be unpredictable.
Please commit and push first.
""", red=True)
                output.annotate(status, red=True)
                raise DeploymentError()
        outgoing, _ = cmd(
            'git log {remote}/{branch}..{branch} --pretty=oneline'.format(
                remote=self.remote, branch=self.branch),
            acceptable_returncodes=[0, 128])
        if outgoing.strip():
            output.error("""\
Your repository has outgoing changes.

I am refusing to deploy in this situation as the results will be unpredictable.
Please push first.
""")
            raise DeploymentError()


class GitPullRepository(GitRepository):

    def _ship(self, host):
        host.rpc.git_pull_code(upstream=self.upstream,
                               branch=self.branch)


class GitBundleRepository(GitRepository):

    def _ship(self, host):
        head = host.rpc.git_current_head(self.branch)
        if head is None:
            bundle_range = self.branch
        else:
            bundle_range = '{head}..{branch}'.format(
                head=head, branch=self.branch)
        fd, bundle_file = tempfile.mkstemp()
        os.close(fd)
        out, err = cmd('git bundle create {file} {range}'.format(
            file=bundle_file, range=bundle_range),
            acceptable_returncodes=[0, 128])
        if 'create empty bundle' in err:
            return
        change_size = os.stat(bundle_file).st_size
        output.annotate(
            'Sending {} bytes of changes'.format(change_size), debug=True)
        rsync = execnet.RSync(bundle_file, verbose=False)
        rsync.add_target(host.gateway,
                         host.remote_repository + '/batou-bundle.git')
        rsync.send()
        os.unlink(bundle_file)
        output.annotate(
            'Unbundling changes', debug=True)
        host.rpc.git_unbundle_code()



"""Secure handling of secrets for deployments.

Each deployment environment has an associated set of secrets (ssh keys,
database credentials, etc.). These are stored in encrypted form, in
secrets/<environment>.cfg.pgp Each deployment environment needs its own secrets
file. During the deployment process, the `secrets` feature decrypts the files
and provides the secrets as an additional overlay to top-level components.

The `secretsedit` program is provided to edit the secrets file in a
convenient fashion::

    secretsedit secrets/myenv.cfg

You need to have gpg configured with a default key and need to have all
involved users' keys for an environment to encrypt again.

`secretsedit` launches $EDITOR with the decrypted secrets file
and encrypts it afterwards. If the secrets file did not exist before, a
new one is created.

"""

from batou import SuperfluousSecretsSection
from .encryption import EncryptedConfigFile
import os.path


def add_secrets_to_environment_override(environment):
    secrets_file = 'secrets/{}.cfg'.format(environment.name)
    if not os.path.exists(secrets_file):
        return
    with EncryptedConfigFile(secrets_file) as f:
        f.read()
        for section_ in f.config.sections():
            if section_ == 'batou':
                continue
            section = section_.replace('component:', '')
            if section not in environment.components:
                environment.exceptions.append(
                    SuperfluousSecretsSection(section))
            o = environment.overrides.setdefault(section, {})
            o.update(f.config.items(section_))
